#!/usr/bin/python
"""
Extract data from UK Met Office Rain Radar NIMROD image files.
Parse NIMROD format image files, display header data and allow extraction of
raster image to an ESRI ASCII (.asc) format file. A bounding box may be
specified to clip the image to the area of interest. Can be imported as a
Python module or run directly as a command line script.
Author: Richard Thomas
Version: 2.00 (27 November 2021)
Public Repository: https://github.com/richard-thomas/MetOffice_NIMROD
Command line usage:
  python nimrod.py [-h] [-q] [-x] [-bbox XMIN XMAX YMIN YMAX] [infile] [outfile]
positional arguments:
  infile                (Uncompressed) NIMROD input filename
  outfile               Output raster filename (*.asc)
optional arguments:
  -h, --help            show this help message and exit
  -q, --query           Display metadata
  -x, --extract         Extract raster file in ASC format
  -bbox XMIN XMAX YMIN YMAX
                        Bounding box to clip raster data to
Note that any bounding box must be specified in the same units and projection
as the input file. The bounding box does not need to be contained by the input
raster but must intersect it.
Example command line usage:
  python nimrod.py -bbox 279906 285444 283130 290440
    -xq 200802252000_nimrod_ng_radar_rainrate_composite_1km_merged_UK_zip
    plynlimon_catchments_rainfall.asc
Example Python module usage:
    import nimrod
    a = nimrod.Nimrod(open(
        '200802252000_nimrod_ng_radar_rainrate_composite_1km_merged_UK_zip',
        'rb'))
    a.query()
    a.extract_asc(open('full_raster.asc', 'w'))
    a.apply_bbox(279906, 285444, 283130, 290440)
    a.query()
    a.extract_asc(open('clipped_raster.asc', 'w'))
Notes:
  1. Valid for v1.7 and v2.6-4 of NIMROD file specification
  2. Assumes image origin is top left (i.e. that header[24] = 0)
  3. Tested on UK and European composite 1km and 5km data, using Python 3.9
     in Windows 10
Copyright (c) 2021 Richard Thomas
(Nimrod.__init__() method based on read_nimrod.py by Charles Kilburn Aug 2008)
This program is free software: you can redistribute it and/or modify
it under the terms of the Artistic License 2.0 as published by the
Open Source Initiative (http://opensource.org/licenses/Artistic-2.0)
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
"""
import array
import dataclasses
import gzip
import itertools
import logging
import os
import pathlib
import re
import struct
import typing as t
import warnings
from typing import Optional

import a2.dataset
import a2.utils
import numpy as np
import pandas as pd

xarray, xarray_dataset_type = a2.utils.utils._import_xarray_and_define_xarray_type(__file__)

TYPE_KEY_TWEETS = object


@dataclasses.dataclass(frozen=True)
class KeyRadar:
    """Stores most common names for keys in radar netcdf datasets"""

    time: str = "time"
    tp: str = "tp_mm"
    longitude: str = "longitude"
    latitude: str = "latitude"
    x_ngt: str = "x_ngt"
    y_ngt: str = "y_ngt"


def get_key_radar(key_radar: t.Optional[KeyRadar], cumulative_radar: bool = False) -> KeyRadar:
    """Get version of KeyRadar depending on setup"""
    if key_radar is None:
        if cumulative_radar:
            key_radar = a2.dataset.radar.KeyRadar(tp="tp_mm_cum")
        else:
            key_radar = a2.dataset.radar.KeyRadar()
    return key_radar


class Nimrod:
    """Reading, querying and processing of NIMROD format rainfall data files."""

    class RecordLenError(Exception):
        """
        Exception Type: NIMROD record length read from file not as expected.
        """

        def __init__(self, actual: int, expected: int, location: str):
            self.message = "Incorrect record length %d bytes (expected %d) at %s." % (actual, expected, location)

    class HeaderReadError(Exception):
        """Exception Type: Read error whilst parsing NIMROD header elements."""

        pass

    class PayloadReadError(Exception):
        """Exception Type: Read error whilst parsing NIMROD raster data."""

        pass

    class BboxRangeError(Exception):
        """
        Exception Type: Bounding box specified out of range of raster image.
        """

        pass

    def __init__(
        self,
        filename: Optional[pathlib.Path | str] = None,
        filename_zipped: Optional[pathlib.Path | str] = None,
        file_content: Optional[t.TextIO] = None,
        fake_data: bool = False,
    ):
        """
        Parse all header and data info from a NIMROD data file into this object.
        (This method based on read_nimrod.py by Charles Kilburn Aug 2008)
        Args:
            filename: filename of NIMROD binary file
        Raises:
            RecordLenError: NIMROD record length read from file not as expected
            HeaderReadError: Read error whilst parsing NIMROD header elements
            PayloadReadError: Read error whilst parsing NIMROD raster data
        """
        self.convert_from_ngt_to_latlong = a2.dataset.units.ConvFromNGToLatLong()
        self.filename_nimrod = filename

        def check_record_len(file: t.TextIO, expected: int, location: str):
            """
            Check record length in C struct is as expected.
            Args:
                file: file to read from
                expected: expected value of record length read
                location: description of position in file (for reporting)
            Raises:
                HeaderReadError: Read error whilst reading record length
                RecordLenError: Unexpected NIMROD record length read from file
            """

            # Unpack length from C struct (Big Endian, 4-byte long)
            try:
                (record_length,) = struct.unpack(">l", file.read(4))
            except Exception:
                raise Nimrod.HeaderReadError
            if record_length != expected:
                raise Nimrod.RecordLenError(record_length, expected, location)

        if filename is not None:
            file = open(filename, "rb")
        elif filename_zipped is not None:
            file = gzip.open(filename_zipped, "rb")
        else:
            file = file_content
        # Header should always be a fixed length record
        check_record_len(file, 512, "header start")

        try:
            # Read first 31 2-byte integers (header fields 1-31)
            gen_ints = array.array("h")
            gen_ints.fromfile(file, 31)
            gen_ints.byteswap()

            # Read next 28 4-byte floats (header fields 32-59)
            gen_reals = array.array("f")
            gen_reals.fromfile(file, 28)
            gen_reals.byteswap()

            # Read next 45 4-byte floats (header fields 60-104)
            spec_reals = array.array("f")
            spec_reals.fromfile(file, 45)
            spec_reals.byteswap()

            # Read next 56 characters (header fields 105-107)
            characters = array.array("B")
            characters.fromfile(file, 56)

            # Read next 51 2-byte integers (header fields 108-)
            spec_ints = array.array("h")
            spec_ints.fromfile(file, 51)
            spec_ints.byteswap()

        except Exception:
            file.close()
            raise Nimrod.HeaderReadError

        check_record_len(file, 512, "header end")

        # Extract strings and make duplicate entries to give meaningful names
        chars = characters.tobytes().decode()
        self.units = chars[0:8]
        self.data_source = chars[8:32]
        self.title = chars[32:55]

        # Store header values in a list so they can be indexed by "element
        # number" shown in NIMROD specification (starts at 1)
        self.hdr_element = [None]  # Dummy value at element 0
        self.hdr_element.extend(gen_ints)
        self.hdr_element.extend(gen_reals)
        self.hdr_element.extend(spec_reals)
        self.hdr_element.extend([self.units])
        self.hdr_element.extend([self.data_source])
        self.hdr_element.extend([self.title])
        self.hdr_element.extend(spec_ints)

        # related to date and time
        self.validity_year = self.hdr_element[1]
        self.validity_month = self.hdr_element[2]
        self.validity_day = self.hdr_element[3]
        self.validity_hour = self.hdr_element[4]
        self.validity_minute = self.hdr_element[5]
        self.validity_second = self.hdr_element[6]
        self.date_year = self.hdr_element[7]
        self.date_month = self.hdr_element[8]
        self.date_day = self.hdr_element[9]
        self.date_hour = self.hdr_element[10]
        self.date_minute = self.hdr_element[11]

        # Duplicate some of values to give more meaningful names
        self.nrows = self.hdr_element[16]
        self.ncols = self.hdr_element[17]
        self.n_data_specific_reals = self.hdr_element[22]
        self.n_data_specific_ints = self.hdr_element[23] + 1
        # Note "+ 1" because header value is count from element 109
        self.y_top = self.hdr_element[34]
        self.y_pixel_size = self.hdr_element[35]
        self.x_left = self.hdr_element[36]
        self.x_pixel_size = self.hdr_element[37]

        # Calculate other image bounds (note these are pixel centres)
        self.x_right = self.x_left + self.x_pixel_size * (self.ncols - 1)
        self.y_bottom = self.y_top - self.y_pixel_size * (self.nrows - 1)

        # Read payload (actual raster data)
        array_size = self.ncols * self.nrows
        check_record_len(file, array_size * 2, "data start")

        self.data = array.array("h")
        try:
            self.data.fromfile(file, array_size)
            self.data.byteswap()
        except Exception:
            file.close()
            raise Nimrod.PayloadReadError
        if fake_data:
            self.data = np.random.randint(0, 100, array_size)
        check_record_len(file, array_size * 2, "data end")
        file.close()

    def query(self):
        """Print complete NIMROD file header information."""

        print("NIMROD file raw header fields listed by element number:")
        print("General (Integer) header entries:")
        for i in range(1, 32):
            print(" ", i, "\t", self.hdr_element[i])
        print("General (Real) header entries:")
        for i in range(32, 60):
            print(" ", i, "\t", self.hdr_element[i])
        print("Data Specific (Real) header entries (%d):" % self.n_data_specific_reals)
        for i in range(60, 60 + self.n_data_specific_reals):
            print(" ", i, "\t", self.hdr_element[i])
        print("Data Specific (Integer) header entries (%d):" % self.n_data_specific_ints)
        for i in range(108, 108 + self.n_data_specific_ints):
            print(" ", i, "\t", self.hdr_element[i])
        print("Character header entries:")
        print("  105 Units:           ", self.units)
        print("  106 Data source:     ", self.data_source)
        print("  107 Title of field:  ", self.title)

        # Print out info & header fields
        # Note that ranges are given to the edge of each pixel
        print(
            "\nValidity Time:  %2.2d:%2.2d on %2.2d/%2.2d/%4.4d"
            % (
                self.hdr_element[4],
                self.hdr_element[5],
                self.hdr_element[3],
                self.hdr_element[2],
                self.hdr_element[1],
            )
        )
        print(
            "Easting range:  %.1f - %.1f (at pixel steps of %.1f)"
            % (
                self.x_left - self.x_pixel_size / 2,
                self.x_right + self.x_pixel_size / 2,
                self.x_pixel_size,
            )
        )
        print(
            "Northing range: %.1f - %.1f (at pixel steps of %.1f)"
            % (
                self.y_bottom - self.y_pixel_size / 2,
                self.y_top + self.y_pixel_size / 2,
                self.y_pixel_size,
            )
        )
        print("Image size: %d rows x %d cols" % (self.nrows, self.ncols))

    def _extract_data_and_header(self):
        """
        Convert data to header in dictionary format and
        data to 2D numpy array
        """

        if self.x_pixel_size != self.y_pixel_size:
            warnings.warn("x_pixel_size(%d) != y_pixel_size(%d)" % (self.x_pixel_size, self.y_pixel_size))

        # Write header to output file. Note that data is valid at the centre
        # of each pixel so "xllcenter" rather than "xllcorner" must be used
        attributes = {}
        attributes["ncols"] = self.ncols
        attributes["nrows"] = self.nrows
        attributes["xllcenter"] = self.x_left
        attributes["yllcenter"] = self.y_bottom
        attributes["cellsize"] = self.y_pixel_size
        attributes["nodata_value"] = self.hdr_element[3]
        attributes["Units"] = self.units
        attributes["Data source"] = self.data_source
        attributes["Title of field"] = self.title

        # Note that ranges are given to the edge of each pixel
        attributes["Validity Time"] = "%2.2d:%2.2d on %2.2d/%2.2d/%4.4d" % (
            self.hdr_element[4],
            self.hdr_element[5],
            self.hdr_element[3],
            self.hdr_element[2],
            self.hdr_element[1],
        )
        attributes["Eastingrange"] = "{:.1f} - {:.1f} (at pixel steps of {:.1f})".format(
            self.x_left - self.x_pixel_size / 2,
            self.x_right + self.x_pixel_size / 2,
            self.x_pixel_size,
        )
        attributes["Northingrange"] = "{:.1f} - {:.1f} (at pixel steps of {:.1f})".format(
            self.y_bottom - self.y_pixel_size / 2,
            self.y_top + self.y_pixel_size / 2,
            self.y_pixel_size,
        )
        attributes["mage size"] = "%d rows x %d cols" % (self.nrows, self.ncols)

        field_values = np.full((self.nrows, self.ncols), np.nan)
        for i in range(self.nrows):
            for j in range(self.ncols):
                field_values[i, j] = self.data[i * self.ncols + j]
        self.attributes = attributes
        self.field_numpy = field_values
        return attributes, field_values

    def _get_coordinates_in_ngt(self):
        """Create coordinates in units of the national grid."""
        self.x_ngt = np.linspace(self.x_left, self.x_right, self.ncols, endpoint=True)
        self.y_ngt = np.linspace(self.y_bottom, self.y_top, self.nrows, endpoint=True)
        if not np.all(np.diff(self.x_ngt) == self.x_pixel_size) or not np.all(np.diff(self.y_ngt) == self.y_pixel_size):
            raise ValueError(
                "Coordinates not spaced as expected: "
                f"{np.all(np.diff(self.x_ngt) == self.x_pixel_size)=} and/or"
                f"{np.all(np.diff(self.y_ngt) == self.y_pixel_size)=}!"
            )
        self.x_ngt, self.y_ngt = np.meshgrid(self.x_ngt, self.y_ngt)
        return self.x_ngt, self.y_ngt

    def _get_coordinates_in_lat_long(self):
        """Create coordinates in units latitude and longitude."""
        self.x_ngt, self.y_ngt = self._get_coordinates_in_ngt()
        # NOTE, order switches from east-west, north-south -> north-south, east-west!
        self.y_lat, self.x_long = self.convert_from_ngt_to_latlong.transform(self.x_ngt, self.y_ngt)

    def extract_ngt_and_tp_mm(self):
        """Create coordinates [national grid] and total precipitation."""
        self._extract_data_and_header()
        self._get_coordinates_in_ngt()
        self._compute_tp_mm()
        return self.x_ngt, self.y_ngt, self.field_tp_mm

    def extract_xarray_dataset(self, version: str = "default"):
        """Convert nimrod file to netcdf file."""
        self._extract_data_and_header()
        if version == "default":
            self._get_coordinates_in_lat_long()
        elif version == "minimal":
            self._get_coordinates_in_ngt()
        else:
            raise NotImplementedError(f"{version=} not implemented!")
        self._compute_tp_mm()
        if version == "default":
            self.ds = xarray.Dataset(
                data_vars=dict(
                    tp_mm_hr_Div32=(
                        ["time", "y", "x"],
                        np.expand_dims(self.field_numpy, 0),
                    ),
                    tp_mm=(["time", "y", "x"], np.expand_dims(self.field_tp_mm, 0)),
                ),
                coords=dict(
                    time=np.expand_dims(
                        np.datetime64(
                            f"{self.date_year:04}-{self.date_month:02}-{self.date_day:02}"
                            f"T{self.date_hour:02}:{self.date_minute:02}:00.000000000"
                        ),
                        0,
                    ),
                    latitude=(["y", "x"], self.y_lat),
                    longitude=(["y", "x"], self.x_long),
                    x_ngt=(["y", "x"], self.x_ngt),
                    y_ngt=(["y", "x"], self.y_ngt),
                ),
                attrs=self.attributes,
            )
        elif version == "minimal":
            self.ds = xarray.Dataset(
                data_vars=dict(
                    tp_mm=(["time", "y", "x"], np.expand_dims(self.field_tp_mm, 0)),
                ),
                coords=dict(
                    time=np.expand_dims(
                        np.datetime64(
                            f"{self.date_year:04}-{self.date_month:02}-{self.date_day:02}"
                            f"T{self.date_hour:02}:{self.date_minute:02}:00.000000000"
                        ),
                        0,
                    ),
                    x_ngt=(["y", "x"], self.x_ngt),
                    y_ngt=(["y", "x"], self.y_ngt),
                ),
                attrs=self.attributes,
            )
        return self.ds

    def _compute_tp_mm(self):
        """Convert total precipitation from code units to total precipitation
        accumulated within time range of file (5 min) in mm."""
        self.field_tp_mm = np.full_like(self.field_numpy, np.nan)
        mask = self.field_numpy != -1
        # data files contain integer precipitation rates in unit of (mm/hr)*32
        # convert from code to (mm) within 5 minutes (time interval of files)
        self.field_tp_mm[mask] = self.field_numpy[mask] / 32 * 5.0 / 60.0

    def to_netcdf(self, filename: str | pathlib.Path):
        """
        Write netcdf file to be read with xarray
        Args:
            filename: name of file data will be save to
        """
        self.extract_xarray_dataset().to_netcdf(filename)


def _convert_file_to_xarray_dataset(
    filename_nimrod: Optional[str | pathlib.Path] = None, file_content_nimrod: Optional[t.TextIO] = None
) -> xarray_dataset_type:
    """Create xarray dataset from nimrod file"""
    nim = Nimrod(filename=filename_nimrod, file_content=file_content_nimrod)
    return nim.extract_xarray_dataset()


def _convert_file(filename: Optional[str | pathlib.Path] = None):
    """Create xarray dataset from zipped nimrod file"""
    with gzip.open(filename, "rb") as file_content:
        try:
            return _convert_file_to_xarray_dataset(file_content_nimrod=file_content)
        except Exception as e:
            warnings.warn(f"Couldn't open nimrod file {filename}!\n{e}")
            return None


def convert_daily_files_to_multiple_netcdf(
    file_list: list[str | pathlib.Path], file_name_netcdf: str | pathlib.Path, chunk_size: int = 20
):
    """
    Convert list of daily files to multiple netcdf files

    Parameters:
    ----------
    file_list: List of daily filenames (nimrod files)
    file_name_netcdf: Prefix for filename of newly created netcdf file
    chunk_size: Number of files per chunk

    Returns
    -------
    """

    # sort file list as otherwise subsequent merge throws:
    # ValueError: Resulting object does not have monotonic global indexes along dimension time
    def match_date_rain_radar_file(filename):
        try:
            return int(
                re.findall(
                    r"^metoffice-c-band-rain-radar_uk_(\d+)_1km-composite.dat.gz",
                    filename,
                )[0]
            )
        except IndexError as e:
            raise ValueError(f"Couldn't parse: {filename}!") from e

    file_list = a2.utils.file_handling.sort_filenames_by_pattern(file_list, match_date_rain_radar_file)
    for i, partial_files in enumerate(a2.utils.utils.chunks(file_list, chunk_size)):
        ds_to_merge = a2.utils.utils.parallelize(_convert_file, partial_files, single_arg=True, processes=1)
        ds_to_merge = [x for x in ds_to_merge if x is not None]
        if len(ds_to_merge):
            xarray.merge(ds_to_merge).to_netcdf(file_name_netcdf.with_suffix(f".{i}.nc"))
        else:
            warnings.warn(f"No netcdf files {ds_to_merge=} found to merge in {partial_files=}!")


def _nimrod_daily_tarball_to_netcdf(filename_tarball: str | pathlib.Path, chunk_size: int = 10) -> pathlib.Path:
    """
    Convert files in single tarball (files for day each) to netcdf

    Parameters:
    ----------
    filename_tarball: Filename of (daily) tarball
    chunk_size: Number of files per chunk saved to single netcdf

    Returns
    -------
    """
    filename_tarball = pathlib.Path(filename_tarball)
    folder_minute_data = filename_tarball.with_suffix("")
    os.makedirs(folder_minute_data, exist_ok=True)
    a2.utils.file_handling.decompress_tarball(filename_tarball, folder_minute_data)
    file_name_netcdf = folder_minute_data / (folder_minute_data.name + ".nc")
    convert_daily_files_to_multiple_netcdf(
        a2.utils.file_handling.get_all_files(folder_minute_data / "*.dat.gz"),
        file_name_netcdf,
        chunk_size=chunk_size,
    )
    return file_name_netcdf


def all_nimrod_daily_tarball_to_netcdf(
    base_folder: t.Optional[str | pathlib.Path] = None,
    file_list: t.Optional[list[str | pathlib.Path]] = None,
    chunk_size: int = 10,
    processes: int = -1,
) -> list[str | pathlib.Path]:
    """
    Convert list of tarballs to netcdf files

    Parameters:
    ----------
    base_folder: Base folder to look for tarballs including its subdirectories
    file_list: List of tarballs of nimrod files
    chunk_size: Number of files per chunk saved to single netcdf
    processes: Number of tarballs saved to netcdf in parallel

    Returns
    -------
    """
    if file_list is None and base_folder is not None:
        file_list = a2.utils.file_handling.get_all_files(
            f"{base_folder.__str__()}/*dat.gz.tar",
            include_subdirectories=True,
        )
    logging.info(f"... converting {len(file_list)} files to netcdf")
    print(f"converting: {file_list=}")
    folder_list = a2.utils.utils.parallelize(
        _nimrod_daily_tarball_to_netcdf,
        zip(file_list),
        kwargs_as_dict={"chunk_size": chunk_size},
        processes=processes,
    )
    return folder_list


def _combine_netcdf(folder_list: list[str | pathlib.Path], filename: str | pathlib.Path) -> bool:
    """Save list of netcdf files to single file and return success status"""
    success = True
    print(f"_combine_netcdf: merging {folder_list=} to {filename=}")
    try:
        a2.dataset.load_dataset.save_dataset(
            xarray.open_mfdataset(folder_list, combine_attrs="drop_conflicts"), filename
        )
    except ValueError as e:
        warnings.warn(f"Couldn't combine netcdf files {folder_list} to single netcdf file {filename}!\n{e}")
        success = False
    return success


def _get_netcdf_filenames_from_tarball_folders(tarball_folders: list[str]) -> list[str]:
    """Create filenames of netcdf files per tarball in list of tarball folder"""
    folders_daily = []
    for tar in tarball_folders:
        folders_daily.append(tar.replace(".tar", ""))
    nc_files_per_folder = []
    for folder in folders_daily:
        match_files = os.path.join(folder, os.path.split(folder)[1] + ".[0-9]*.nc")
        files_in_folder = a2.utils.file_handling.get_all_files(match_files)
        nc_files_per_folder.append(files_in_folder)
    return nc_files_per_folder


def merge_nc_files_on_daily_basis(
    base_folder: str, input_is_folders: bool = False, remove_files: bool = False, processes: int = -1
):
    """
    Merge all netcdf files in subfolders of base_folder to single netcdf per day.
    This assumes the internal filename convention for nimrod filenames and
    that all all netcdf filenames describing a single day are found in a single folder.

    Parameters:
    ----------
    base_folder: Subfolders are searched for pattern "**/*.dat.gz.tar" following naming convention,
                 alternatively subfolders can be given see `input_is_folders`
    input_is_folders: Is `base_folder` a list of folders containing netcdf files to be merged
    remove_files: Single files deleted after merging
    processes: Number of files merged to single file in parallel

    Returns
    -------
    """
    if not input_is_folders:
        tarballs = a2.utils.file_handling.get_all_files(
            os.path.join(base_folder.__str__(), "**/*.dat.gz.tar"), recursive=True
        )
    else:
        tarballs = base_folder
    nc_files_per_folder = _get_netcdf_filenames_from_tarball_folders(tarballs)
    print(f"{nc_files_per_folder=}")
    a2.utils.utils.parallelize(
        _merge_nc_files_on_daily_basis,
        zip(nc_files_per_folder),
        single_arg=True,
        kwargs_as_dict={"remove_files": remove_files},
        processes=processes,
    )


def _merge_nc_files_on_daily_basis(files, remove_files=False):
    """
    See `merge_nc_files_on_daily_basis`. Function to be parallelized.
    """
    if not len(files):
        logging.info(f"Nothing to merge {files=}")
        return
    filename_netcdf = re.sub(r".dat.gz.\d+.nc", r".dat.gz.nc", files[0])
    success = _combine_netcdf(files, filename_netcdf)
    logging.info(f"merging files into {filename_netcdf}")
    if success and remove_files:
        for f in files:
            os.remove(f)


def from_ngt_to_ds_index(
    ds_radar: xarray_dataset_type,
    x: np.ndarray,
    y: np.ndarray,
    x_key: str = "x_ngt",
    y_key: str = "y_ngt",
    asserting: bool = False,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """
    Convert x and y values (in ngt) to indices in `ds_radar` of respective fields

    Parameters:
    ----------
    ds_radar: Radar data as xarray dataset
    x: Values along west-east in national grid units
    y: Values along north-south in national grid units
    x_key: Key in dataset to corresponding west-east coordinate values in national grid units
    y_key: Key in dataset to corresponding north-south coordinate values in national grid units
    asserting: Additional (costly) consistency checks are done

    Returns
    indices for x, indices for y
    -------
    """
    y_shape, x_shape = ds_radar[x_key].shape
    x_min = ds_radar[x_key].min().item()
    x_max = ds_radar[x_key].max().item()
    x_spacing = (x_max - x_min) / (x_shape - 1)

    y_min = ds_radar[y_key].min().item()
    y_max = ds_radar[y_key].max().item()
    y_spacing = (y_max - y_min) / (y_shape - 1)

    x_index = (x - x_min) / x_spacing
    y_index = (y - y_min) / y_spacing
    x_index, y_index = x_index.astype(int), y_index.astype(int)

    if asserting:
        ds_test = ds_radar.sel(x=x_index, y=y_index)
        unique_values_x = np.sort(np.unique(ds_test[x_key].values))
        unique_values_y = np.sort(np.unique(ds_test[y_key].values))
        x_unique = np.sort(np.unique(x))
        y_unique = np.sort(np.unique(y))
        if not np.array_equal(unique_values_x, x_unique) or not np.array_equal(unique_values_y, y_unique):
            for u, s in zip(unique_values_x, x_unique):
                print(f"{u=}, {s=}")
            for u, s in zip(unique_values_y, y_unique):
                print(f"{u=}, {s=}")
            raise ValueError(
                f"Indices not correctly identified: {unique_values_x=} "
                f"should be {x_unique=} and {unique_values_y=} should be {y_unique=}"
            )
    return x_index, y_index


def get_time_ds_index(
    ds_radar: xarray_dataset_type, times: list, key_time: str = "time", check_for_outliers: bool = True
) -> np.ndarray:
    """
    Convert date time values to indices in `ds_radar` of respective time field.

    Parameters:
    ----------
    ds_radar: Radar data as xarray dataset
    times: Time values
    key_time: Key in dataset to corresponding date time
    check_for_outliers: Additional consistency checks are done

    Returns
    indices for x, indices for y
    -------
    """
    times_int = times.astype(int)
    times_ds_int = ds_radar[key_time].values.astype(int)
    time_shape = times_ds_int.shape[0]
    if time_shape == 1:
        if not all(ds_radar[key_time] == t for t in times):
            raise ValueError(
                f"Single time present in dataset {ds_radar[key_time]} doesn't correspond to desired value {times}!"
            )
        else:
            return np.zeros(len(times), dtype=int)
    time_min = times_ds_int.min()
    time_max = times_ds_int.max()
    time_spacing = (time_max - time_min) / (time_shape - 1)
    mask = np.logical_and(times_int >= time_min, times_int <= time_max)
    if check_for_outliers and times_int[~mask].shape[0]:
        raise ValueError(f"Found times outide of dataset range: {times_int[~mask]=},{time_min=},{time_max=}")
    return ((times_int - time_min) / time_spacing).astype(int)


def assign_ngt_to_tweets(
    ds_tweets: xarray_dataset_type,
    key_tweets=None,
) -> xarray_dataset_type:
    """Compute coordinates in units of the national grid from latitude and longitude
    and save them to the dataset"""
    ds_tweets = ds_tweets.copy()
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    ds_tweets = a2.dataset.load_dataset.reset_index_coordinate(ds_tweets)
    converter_from_latong_to_ngt = a2.dataset.units.ConvFromLatLongToNG()
    results = a2.utils.utils.parallelize(
        converter_from_latong_to_ngt.transform,
        zip(
            ds_tweets[key_tweets.latitude].values,
            ds_tweets[key_tweets.longitude].values,
        ),
    )
    easting, northing = np.array(results).T
    ds_tweets[key_tweets.x_ngt] = (["index"], easting)
    ds_tweets[key_tweets.y_ngt] = (["index"], northing)
    return ds_tweets


def assign_radar_to_tweets_from_netcdf(
    ds_tweets: xarray_dataset_type,
    list_radar_filenames: t.Optional[list[str | pathlib.Path]] = None,
    base_folder: t.Optional[str | pathlib.Path] = None,
    key_tweets: TYPE_KEY_TWEETS | None = None,
    key_radar: KeyRadar = KeyRadar(),
    round_ngt_offset: int = 500,
    round_ngt_decimal: int = -3,
    round_time_to_base: int = 5,
) -> xarray_dataset_type:
    """
    Assign radar precipitation data that is stored in netcdf files to Tweets dataset.
    Tweets coordinates (latitude, longitude, time) are converted
    and rounded to value convention in radar dataset.

    Parameters:
    ----------
    ds_tweets: Dataset of Tweets
    list_radar_filenames: List of netcdf files storing radar data
    base_folder: Base folder to look for to look for netcdf files storing radar data (*.nc)
    key_tweets: Dataclass storing key convention for tweets datasets
    key_radar: Dataclass storing key convention for radar datasets
    round_ngt_offset: Round spatial coordinates of Tweets to this offset
    round_ngt_decimal: Round spatial coordinates of Tweets to decimal point
    round_time_to_base: Round time coordinate of Tweets to this base

    Returns
    dataset with added radar data
    -------
    """
    if not list_radar_filenames:
        radar_filenames = a2.utils.file_handling.get_all_files(
            os.path.join(base_folder.__str__(), "**/*.nc"), recursive=True
        )
    else:
        radar_filenames = list_radar_filenames
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    ds_tweets = _prepare_tweet_ds_for_radar(
        ds_tweets,
        key_tweets=key_tweets,
        round_ngt_offset=round_ngt_offset,
        round_ngt_decimal=round_ngt_decimal,
        round_time_to_base=round_time_to_base,
    )

    ds_tweets[key_tweets.tp_radar] = (
        ["index"],
        np.full(ds_tweets.index.shape[0], np.nan),
    )
    for radar_filename in radar_filenames:
        ds_radar = xarray.open_dataset(radar_filename)
        print(f"{ds_radar=}")
        radar_times = ds_radar[key_radar.time].values
        radar_times_min = radar_times.min()
        radar_times_max = radar_times.max()
        mask = (ds_tweets[key_tweets.time_radar] >= radar_times_min) & (
            ds_tweets[key_tweets.time_radar] <= radar_times_max
        )
        print(f"{radar_times=}")
        indices_select = ds_tweets["index"].loc[mask]

        tweets_x_ngt_rounded = ds_tweets[key_tweets.x_ngt_rounded].sel(index=indices_select).values
        tweets_y_ngt_rounded = ds_tweets[key_tweets.y_ngt_rounded].sel(index=indices_select).values
        time = ds_tweets[key_tweets.time_radar].sel(index=indices_select).values

        radar_values_sel = _get_tp_from_ds_radar(key_radar, ds_radar, tweets_x_ngt_rounded, tweets_y_ngt_rounded, time)
        ds_tweets[key_tweets.tp_radar].loc[dict(index=indices_select)] = radar_values_sel
    return ds_tweets


def _get_tp_from_ds_radar(
    key_radar: KeyRadar,
    ds_radar: xarray_dataset_type,
    tweets_x_ngt_rounded: np.ndarray,
    tweets_y_ngt_rounded: np.ndarray,
    time: np.datetime64,
    asserting_xy: bool = False,
) -> np.ndarray:
    """
    Based on spatial coordinates (in units of the national grid) and time obtain total precipitation from radar dataset.

    Parameters:
    ----------
    key_radar: Dataclass storing key convention for radar datasets
    ds_radar: Dataset of Radar data
    tweets_x_ngt_rounded: Coordinates along west-east in units of the national grid for the Tweets
    tweets_y_ngt_rounded: Coordinates along north-south in units of the national grid for the Tweets
    time: Time values for Tweets
    asserting_xy: Additional (costly) consistency checks on spatial coordinates indices are done

    Returns
    -------
    radar values
    """
    index_x, index_y = from_ngt_to_ds_index(
        ds_radar, tweets_x_ngt_rounded, tweets_y_ngt_rounded, asserting=asserting_xy
    )
    index_time = get_time_ds_index(ds_radar, time, key_time=key_radar.time)
    dims_required = ("time", "y", "x")
    dims_found = ds_radar[key_radar.tp].coords.dims
    if dims_found != dims_required:
        raise IndexError(
            f"Require dims of precipitation {key_radar.tp} in radar ds to be {dims_required} but found {dims_found}"
        )
    radar_values_sel = ds_radar[key_radar.tp].values[index_time, index_y, index_x]
    return radar_values_sel


def _prepare_tweet_ds_for_radar(
    ds_tweets: xarray_dataset_type,
    key_tweets: Optional[TYPE_KEY_TWEETS] = None,
    round_ngt_offset: int = 500,
    round_ngt_decimal: int = -3,
    round_time_to_base: int = 5,
) -> xarray_dataset_type:
    """
    Prepare Tweets dataset by converting latitude, longitude to
    national grid units at consistent intervals with radio data.
    In addition, round time to same convention as radar data.

    Parameters:
    ----------
    ds_tweets: Dataset of Tweets
    key_tweets: Dataclass storing key convention for tweets datasets
    round_ngt_offset: Round spatial coordinates of Tweets to this offset
    round_ngt_decimal: Round spatial coordinates of Tweets to decimal point
    round_time_to_base: Round time coordinate of Tweets to this base

    Returns
    dataset with prepared coordinates for radar data assignment
    -------
    """
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    ds_tweets = assign_ngt_to_tweets(
        ds_tweets,
        key_tweets=key_tweets,
    )
    ds_tweets[key_tweets.x_ngt_rounded] = (
        ["index"],
        a2.utils.utils.round_offset(
            ds_tweets[key_tweets.x_ngt].values,
            decimal=round_ngt_decimal,
            offset=round_ngt_offset,
        ),
    )
    ds_tweets[key_tweets.y_ngt_rounded] = (
        ["index"],
        a2.utils.utils.round_offset(
            ds_tweets[key_tweets.y_ngt].values,
            decimal=round_ngt_decimal,
            offset=round_ngt_offset,
        ),
    )
    ds_tweets[key_tweets.time_radar] = (
        ["index"],
        a2.utils.utils.vetorize_time_to_base_minutes(ds_tweets[key_tweets.time].values, base=round_time_to_base),
    )

    return ds_tweets


def assign_radar_to_tweets(
    ds_tweets: xarray_dataset_type,
    key_tweets: Optional[TYPE_KEY_TWEETS] = None,
    round_ngt_offset: int = 500,
    round_ngt_decimal: int = -3,
    round_time_to_base: int = 5,
    path_to_dapceda: str | pathlib.Path | None = None,
    processes: int = -1,
) -> xarray_dataset_type:
    """
    Assign radar data from raw radar data files to Tweets dataset.
    Radar files are unzipped and preprocessed for this.

    Parameters:
    ----------
    ds_tweets: Dataset of Tweets
    key_tweets: Dataclass storing key convention for tweets datasets
    round_ngt_offset: Round spatial coordinates of Tweets to this offset
    round_ngt_decimal: Round spatial coordinates of Tweets to decimal point
    round_time_to_base: Round time coordinate of Tweets to this base
    path_to_dapceda: Path to "badc" folder downloaded as downloaded form CEDA archive
    processes: Number of radar files assigned to Tweets in parallel

    Returns
    dataset with added radar data
    -------
    """
    if key_tweets is None:
        key_tweets = a2.dataset.tweets.KeyTweets()
    if key_tweets is None:
        key_tweets = key_tweets()
    if path_to_dapceda is None:
        path_to_dapceda = a2.utils.file_handling.get_folder_radar()
    path_to_dapceda = pathlib.Path(path_to_dapceda)
    if not path_to_dapceda.is_dir():
        raise ValueError(f"Couldn't find folder {path_to_dapceda=}!")
    ds_tweets = _prepare_tweet_ds_for_radar(
        ds_tweets,
        key_tweets=key_tweets,
        round_ngt_offset=round_ngt_offset,
        round_ngt_decimal=round_ngt_decimal,
        round_time_to_base=round_time_to_base,
    )

    ds_tweets[key_tweets.tp_radar] = (
        ["index"],
        np.full(ds_tweets.index.shape[0], np.nan),
    )
    ds_tweets[key_tweets.time_radar_int] = (["index"], ds_tweets[key_tweets.time_radar].values.astype(int))
    ds_tweets, time_xngt_yngt = a2.dataset.utils_dataset.divide_ds_by_unique_values(
        ds_tweets,
        key_divide_by=key_tweets.time_radar_int,
        keys_values=[key_tweets.time_radar, key_tweets.x_ngt_rounded, key_tweets.y_ngt_rounded],
    )

    radar_tp_mm = a2.utils.utils.parallelize(
        get_tp_from_single_ds_radar,
        zip(*time_xngt_yngt),
        single_arg=False,
        kwargs_as_dict={
            "path_to_dapceda": path_to_dapceda,
        },
        processes=processes,
    )

    radar_tp_mm = np.array(a2.utils.utils.flatten_list(radar_tp_mm))
    ds_tweets[key_tweets.tp_radar] = (["index"], radar_tp_mm)
    return ds_tweets


def get_tp_from_single_ds_radar(
    time,
    tweets_x_ngt_rounded,
    tweets_y_ngt_rounded,
    path_to_dapceda=None,
    key_radar=KeyRadar(),
) -> np.ndarray:
    """
    Retrieve precipitation from single radar files at various locations.
    Finds respective binary file for time, converts it to an xarray dataset and returns precipitation
    at the desired locations.

    Parameters:
    ----------
    time: Time stamps for the desired coordinates (should all be same)
    tweets_x_ngt_rounded: Coordinates along west-east in units of the national grid for the Tweets
    tweets_y_ngt_rounded: Coordinates along north-south in units of the national grid for the Tweets
    path_to_dapceda: Path to "badc" folder downloaded as downloaded form CEDA archive
    key_radar: Dataclass storing key convention for radar datasets

    Returns
    radar values at given locations
    -------
    """
    if not all(time[0] == x for x in time):
        raise Exception(f"Time data not same time {time=}!")
    radar_filename_zipped, filename_tarball, folder_to_decompress = _get_filename_from_time(path_to_dapceda, time[0])
    if not radar_filename_zipped.is_file() and not folder_to_decompress.is_dir() and filename_tarball.is_file():
        a2.utils.file_handling.decompress_tarball(filename_tarball, folder_to_decompress)
    if not radar_filename_zipped.is_file():
        warnings.warn(f"Cannot find file {radar_filename_zipped}, skipping it ....")
        return np.full(np.shape(time), -np.inf)
    try:
        nim = a2.dataset.radar.Nimrod(filename_zipped=radar_filename_zipped)
    except (Nimrod.RecordLenError, Nimrod.HeaderReadError, Nimrod.PayloadReadError, Nimrod.BboxRangeError) as e:
        warnings.warn(f"{e}: Cannot convert file {radar_filename_zipped=}, skipping it ....")
        return np.full(np.shape(time), -np.inf)

    ds_radar = nim.extract_xarray_dataset(version="minimal")
    radar_values_sel = _get_tp_from_ds_radar(
        key_radar,
        ds_radar,
        np.array(tweets_x_ngt_rounded),
        np.array(tweets_y_ngt_rounded),
        np.array(time),
    )
    return radar_values_sel


def _get_filename_from_time(
    path_to_dapceda: pathlib.Path, time: np.datetime64
) -> t.Tuple[pathlib.Path, pathlib.Path, pathlib.Path]:
    """Determine radar filename, folder based on timestamp"""
    year, month, day, hour, minute = a2.utils.times.time_breakdown(time)
    radar_filename_zipped = (
        path_to_dapceda / f"badc/ukmo-nimrod/data/composite/uk-1km/{year}/"
        f"metoffice-c-band-rain-radar_uk_{year}{month:02}{day:02}_1km-composite.dat.gz/"
        f"metoffice-c-band-rain-radar_uk_{year}{month:02}{day:02}{hour:02}{minute:02}_1km-composite.dat.gz"
    )
    filename_tarball = (
        path_to_dapceda / f"badc/ukmo-nimrod/data/composite/uk-1km/{year}/"
        f"metoffice-c-band-rain-radar_uk_{year}{month:02}{day:02}_1km-composite.dat.gz.tar"
    )
    folder_to_decompress = (
        path_to_dapceda / f"badc/ukmo-nimrod/data/composite/uk-1km/{year}/"
        f"metoffice-c-band-rain-radar_uk_{year}{month:02}{day:02}_1km-composite.dat.gz"
    )
    return radar_filename_zipped, filename_tarball, folder_to_decompress


def nimrod_ds_from_time(
    path_to_dapceda: pathlib.Path, time: np.datetime64, version: str = "default"
) -> xarray_dataset_type:
    """Retrieve nimrod file based on time and return it as xarray.Dataset.
    `version` determines format of returned Dataset"""
    path_to_dapceda = pathlib.Path(path_to_dapceda)
    radar_filename_zipped, filename_tarball, folder_to_decompress = _get_filename_from_time(path_to_dapceda, time)
    if not radar_filename_zipped.is_file() and not folder_to_decompress.is_dir() and filename_tarball.is_file():
        a2.utils.file_handling.decompress_tarball(filename_tarball, folder_to_decompress)
    try:
        nim = a2.dataset.radar.Nimrod(filename_zipped=radar_filename_zipped)
    except (
        Nimrod.RecordLenError,
        Nimrod.HeaderReadError,
        Nimrod.PayloadReadError,
        Nimrod.BboxRangeError,
        FileNotFoundError,
    ) as e:
        warnings.warn(f"{e}: Cannot convert file {radar_filename_zipped=}, skipping it ....")
        return None
    ds_radar = nim.extract_xarray_dataset(version=version)
    return ds_radar


def nimrod_ds_cumulative_from_time(
    path_to_dapceda: pathlib.Path,
    time: np.datetime64,
    time_delta: float = 1,
    time_delta_units: str = "h",
    version: str = "default",
) -> xarray_dataset_type:
    """
    Compute total precipitation by summing cumulative over time period `time_delta` [`time_delta_units`] up to `time`.

    Parameters:
    ----------
    path_to_dapceda: Path to "badc" folder downloaded as downloaded form CEDA archive
    time: Precipitation summed up to this time
    time_delta: Precipitation summed over this period of time
    time_delta_units: Units of `time_delta`
    version: Determines format of returned Dataset
             'default': Returns coordinates as longitude/latitude and in northing/easting of national grid
             'minimal': Ony returns coordinates as longitude/latitude

    Returns
    dataset with cumulated precipitation
    -------
    """
    coords = None
    ds_not_none = None
    time_increment_files = pd.Timedelta("5m")
    time_end = pd.to_datetime(time)
    time_start = pd.to_datetime(time) - pd.Timedelta(f"{time_delta}{time_delta_units}") + time_increment_files
    ds = nimrod_ds_from_time(path_to_dapceda, time_start, version=version)
    tp_list = []
    tp_list, coords, ds_not_none = _append_to_list(ds, tp_list, coords, ds_not_none)
    t = time_start + time_increment_files
    while t <= time_end:
        ds = nimrod_ds_from_time(path_to_dapceda, t)
        tp_list, coords, ds_not_none = _append_to_list(ds, tp_list, coords, ds_not_none)
        t = t + time_increment_files
    if not len(tp_list):
        warnings.warn(
            f"Found no files at {time=} with {time_delta=}, "
            f"{time_delta_units=} for cumulative nimrod ds, returning `None`"
        )
        return None
    coords_time = np.expand_dims(
        np.datetime64(time_end),
        0,
    )
    data_vars = dict(
        tp_mm_cum=(["time", "y", "x"], np.sum(tp_list, axis=0)),
    )
    return a2.dataset.utils_dataset.construct_dataset(ds_not_none, data_vars, time=coords_time)


def _append_to_list(ds, tp_list, coords, ds_not_none):
    """Helper function to append list with precipitation values"""
    if ds is not None:
        tp_list.append(ds.tp_mm.values)
        coords = ds.coords
        ds_not_none = ds
    return tp_list, coords, ds_not_none


def time_series_from_files(
    time_start: np.datetime64,
    time_end: np.datetime64,
    longitudes: np.ndarray,
    latitudes: np.ndarray,
    path_to_dapceda: Optional[str | pathlib.Path] = None,
    time_delta: int = 1,
    time_delta_units: str = "h",
    key_radar: Optional[KeyRadar] = None,
    key_time: str = "time",
    processes: int = -1,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """
    Compute time series of cumulative radar precipitation for location given in long/lat.

    Parameters:
    ----------
    time_start: Time series starting time
    time_end: Time series final time
    longitudes: Longitudes of locations of interest
    latitudes: Latitudes of locations of interest
    path_to_dapceda: Path to "badc" folder downloaded as downloaded form CEDA archive
    time_delta: Precipitation summed over this period of time
    time_delta_units: Units of `time_delta`
    key_radar: Dataclass storing key convention for radar datasets
    key_time: Key in dataset to corresponding date time
    processes: Number of times computed in parallel

    Returns
    times, cumulative precipitation with shape: [times, lat/long, 0/1]
    -------
    """
    if path_to_dapceda is None:
        path_to_dapceda = a2.utils.constants.PATH_DAPCEDA
    if key_radar is None:
        key_radar = KeyRadar(tp="tp_mm_cum")
    time_increment = pd.Timedelta(f"{time_delta}{time_delta_units}")
    if not time_increment % pd.Timedelta("5m") == pd.Timedelta(0):
        raise ValueError(f"{time_delta}{time_delta_units} should be divisible by time spacing of files (5 min)!")
    time_end = a2.utils.utils.round_time_to_base_minutes(time_end, base=5)
    time_end = pd.to_datetime(time_end)
    time_start = a2.utils.utils.round_time_to_base_minutes(pd.to_datetime(time_start), base=5)
    time_start = pd.to_datetime(time_start)
    logging.info(
        f"Computing time series from {time_start=} to {time_end=} in increments of {time_delta}{time_delta_units}"
    )
    t = time_start
    converter_from_latong_to_ngt = a2.dataset.units.ConvFromLatLongToNG()
    x_ngt, y_ngt = np.array(converter_from_latong_to_ngt.transform(latitudes, longitudes))
    x_ngt_rounded = a2.utils.utils.round_offset(x_ngt, decimal=-3, offset=500)
    y_ngt_rounded = a2.utils.utils.round_offset(y_ngt, decimal=-3, offset=500)

    times = []
    while t <= time_end:
        times.append(t)
        t = t + time_increment
    times_tp = a2.utils.utils.parallelize(
        _get_values_single_time,
        zip(times, itertools.repeat(x_ngt_rounded), itertools.repeat(y_ngt_rounded)),
        processes=processes,
        single_arg=False,
        kwargs_as_dict=dict(
            path_to_dapceda=path_to_dapceda,
            time_delta=time_delta,
            time_delta_units=time_delta_units,
            key_radar=key_radar,
            key_time=key_time,
        ),
    )
    times_radar = [t[0] for t in times_tp]
    tp_radar = [t[1] for t in times_tp]
    return np.array(times_radar), np.array(tp_radar)


def _get_values_single_time(
    time: np.datetime64,
    x_ngt_rounded: np.ndarray,
    y_ngt_rounded: np.ndarray,
    path_to_dapceda: Optional[str | pathlib.Path] = None,
    time_delta: int = 1,
    time_delta_units: str = "h",
    key_radar: Optional[KeyRadar] = None,
    key_time: str = "time",
) -> t.Tuple[np.datetime64, np.ndarray]:
    """
    Obtains precipitation values from single file for all given locations.

    Parameters:
    ----------
    time: Time to get precipitation values
    x_ngt_rounded: Coordinates along west-east in units of the national grid for the Tweets
    y_ngt_rounded: Coordinates along north-south in units of the national grid for the Tweets
    path_to_dapceda: Path to "badc" folder downloaded as downloaded form CEDA archive
    time_delta: Precipitation summed over this period of time
    time_delta_units: Units of `time_delta`
    key_radar: Dataclass storing key convention for radar datasets
    key_time: Key in dataset to corresponding date time

    Returns
    -------
    time, list of radar precipitation shape: [long/lat]
    """
    if path_to_dapceda is None:
        path_to_dapceda = a2.utils.constants.PATH_DAPCEDA
    ds = nimrod_ds_cumulative_from_time(
        path_to_dapceda, time, time_delta=time_delta, time_delta_units=time_delta_units, version="minimal"
    )
    if ds is None:
        return time, np.full_like(x_ngt_rounded, np.nan)
    tp_list = []
    for x, y in zip(x_ngt_rounded, y_ngt_rounded):
        tp = _get_tp_from_ds_radar(key_radar, ds, x, y, np.array([time], dtype=np.datetime64), asserting_xy=True)
        tp_list.append(tp)
    return ds[key_time].values[0], np.array(tp_list)



import a2.plotting.analysis
import a2.plotting.figures
import a2.plotting.histograms
import a2.plotting.parallel_plotting
import a2.plotting.timeseries
import a2.plotting.utils_plotting
import a2.plotting.weather_maps



import a2.dataset.emojis
import a2.dataset.load_dataset
import a2.dataset.radar
import a2.dataset.stations
import a2.dataset.tweets
import a2.dataset.units
import a2.dataset.utils_dataset
from a2.dataset.utils_dataset import is_nan



import typing as t

import a2.utils.constants
import matplotlib.pyplot as plt


def set_axes(
    ax: plt.axes,
    xlim: t.Optional[list] = None,
    ylim: t.Optional[list] = None,
    fontsize: int = 8,
    title: t.Optional[str] = None,
    label_x: t.Optional[str] = None,
    label_y: t.Optional[str] = None,
    labelrotation_x: t.Optional[float] = None,
    labelrotation_y: t.Optional[float] = None,
):
    """
    Customizes matplotlib axes object

    Parameters:
    ----------
    ax: matplotlib axes
    xlim: Axes limits along the x-axis
    ylim: Axes limits along the y-axis
    fontsize: Size of the font
    title: Title of the axes
    label_x: Label of the x-axis
    label_y: Label of the y-axis
    labelrotation_x: Angle of rotation of the label of the x-axis
    labelrotation_y: Angle of rotation of the label of the y-axis

    Returns
    -------
    """

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    if label_x is not None:
        ax.set_xlabel(label_x, fontsize=fontsize)
    if label_y is not None:
        ax.set_ylabel(label_y, fontsize=fontsize)
    if title is not None:
        ax.set_title(title, fontsize=fontsize)
    ax.tick_params(axis="both", which="major", labelsize=fontsize)
    ax.tick_params(axis="both", which="minor", labelsize=fontsize)
    if labelrotation_x is not None:
        ax.tick_params(axis="x", labelrotation=labelrotation_x)
    if labelrotation_y is not None:
        ax.tick_params(axis="y", labelrotation=labelrotation_y)


def set_colorbar(ax_colorbar: plt.axes, label_y: str, fontsize: float):
    """
    Customizes matplotlib colorbar axes

    Parameters:
    ----------
    ax_colorbar: matplotlib colorbar axes
    label_y: Label of the y-axis
    fontsize: Size of the font

    Returns
    -------
    """
    ax_colorbar.set_ylabel(label_y, fontsize=fontsize)
    ax_colorbar.tick_params(axis="both", which="major", labelsize=fontsize)
    ax_colorbar.tick_params(axis="both", which="minor", labelsize=fontsize)


def set_x_lim(axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES, xlim: tuple | None):
    if xlim is not None:
        axes.set_xlim(*xlim)


def set_y_lim(axes: a2.utils.constants.TYPE_MATPLOTLIB_AXES, ylim: tuple | None):
    if ylim is not None:
        axes.set_ylim(*ylim)



import logging as logging_std_library

import a2.training.benchmarks as timer
import a2.utils.utils
import transformers
from transformers.trainer import *  # Ugly but probably needed ... # noeq

torch = a2.utils.utils._import_torch(__file__)


class TimerCallback(transformers.TrainerCallback):
    """
    # Example usage:
    tmr = timer.CPUGPUTimer()
    trainer = transformers.Trainer(
        # Other args...
        callbacks=[TimerCallback(tmr, gpu=True)]
    )
    trainer.train()
    tmr.print_all_time_stats()
    """

    def __init__(self, timer, gpu=False):
        super().__init__()
        self.timer = timer
        self.gpu = gpu

    def on_epoch_begin(self, args, state, control, **kwargs):
        self.timer.start(timer.TimeType.EPOCH)

    def on_epoch_end(self, args, state, control, **kwargs):
        self.timer.end(timer.TimeType.EPOCH)
        self.timer.complete_all()

    def on_step_begin(self, args, state, control, **kwargs):
        logging_std_library.info(
            f"Epoch {int(state.epoch)}: Start iteration step {state.global_step}/{state.max_steps} of training..."
        )
        self.timer.start(timer.TimeType.BATCH, gpu=self.gpu)

    def on_step_end(self, args, state, control, **kwargs):
        self.timer.end(timer.TimeType.BATCH, gpu=self.gpu)
        if state.global_step % 10 == 0:
            self.timer.complete_all()


class TimeLoaderWrapper:
    """Wrapper around a DataLoader (*not* a Dataset!) for I/O timing."""

    def __init__(self, loader, timer):
        self.loader = loader
        self.tmr = timer

    @staticmethod
    def time_loader(loader, tmr):
        if len(loader) > 0:
            tmr.start(timer.TimeType.IO)
        for i, data in enumerate(loader):
            tmr.end(timer.TimeType.IO)
            if i % 10 == 0:
                tmr.complete_all()
            yield data
            if i != len(loader) - 1:
                tmr.start(timer.TimeType.IO)

    def __iter__(self):
        return TimeLoaderWrapper.time_loader(self.loader, self.tmr)

    def __len__(self):
        return len(self.loader)

    def reset(self):
        if hasattr(self.loader, "reset"):
            self.loader.reset()


class TrainerWithTimer(Trainer):
    """
    Custom Trainer subclass to support finer-grained timing.
    Should also use the callback above.
    Adapted from original:
    # https://github.com/huggingface/transformers/blob/v4.25.1/src/transformers/trainer.py
    """

    def __init__(self, *args, **kwargs):
        print(f'{kwargs["callbacks"]=}')
        self.timer_callback = kwargs["callbacks"][0]
        self.tmr = self.timer_callback.timer
        self.tmr_gpu = self.timer_callback.gpu
        super().__init__(*args, **kwargs)

    def training_step(self, model: nn.Module, inputs: Dict[str, Union[torch.Tensor, Any]]) -> torch.Tensor:
        """
        Perform a training step on a batch of inputs.

        Subclass and override to inject custom behavior.

        Args:
            model (`nn.Module`):
                The model to train.
            inputs (`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument `labels`. Check your model's documentation for all accepted arguments.

        Return:
            `torch.Tensor`: The tensor with training loss on this batch.
        """
        model.train()
        inputs = self._prepare_inputs(inputs)

        if is_sagemaker_mp_enabled():
            loss_mb = smp_forward_backward(model, inputs, self.args.gradient_accumulation_steps)
            return loss_mb.reduce_mean().detach().to(self.args.device)
        if self.tmr:
            self.tmr.start(timer.TimeType.FORWARD, gpu=self.tmr_gpu)

        with self.compute_loss_context_manager():
            loss = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training
        if self.tmr:
            self.tmr.end(timer.TimeType.FORWARD, gpu=self.tmr_gpu)
            self.tmr.start(timer.TimeType.BACKWARD, gpu=self.tmr_gpu)

        if self.use_apex:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            self.accelerator.backward(loss)
        if self.tmr:
            self.tmr.end(timer.TimeType.BACKWARD, gpu=self.tmr_gpu)

        return loss.detach() / self.args.gradient_accumulation_steps

    def get_train_dataloader(self) -> DataLoader:
        dl = super().get_train_dataloader()
        if self.tmr:
            return TimeLoaderWrapper(dl, self.tmr)
        return dl

    def save_model(self, output_dir: Optional[str] = None, _internal_call: bool = False):
        self.tmr.start(timer.TimeType.SAVING_MODEL)
        super().save_model(output_dir=output_dir, _internal_call=_internal_call)
        self.tmr.end(timer.TimeType.SAVING_MODEL)



import dataclasses
import logging
import pprint
import resource
import time

import a2.utils.utils
import numpy as np

torch = a2.utils.utils._import_torch(__file__)


@dataclasses.dataclass(frozen=True)
class TimeType:
    EPOCH: str = "EPOCH"
    BATCH: str = "BATCH"
    FORWARD: str = "FORWARD"
    BACKWARD: str = "BACKWARD"
    IO: str = "IO"
    TRAINING: str = "TRAINING"
    RUN: str = "RUN"
    SAVING_MODEL: str = "SAVING_MODEL"
    EVALUATION: str = "EVALUATION"


MAX_LENGTH_TIME_TYPES = max([len(x) for x in TimeType().__dict__.values()])


def current_time():
    return time.time()


class Timer:
    """Interface has analagous implementation to deep500 timer. Therefore not all variables used"""

    def __init__(self, print_all_single_time_stats=True, debug=True):
        self.times_archive = {}
        self.times_running = {}
        self.TimeType = TimeType()
        self.debug = debug
        self.print_all_single_time_stats = print_all_single_time_stats

    def start(self, time_type, gpu=None):
        if time_type in self.TimeType.__dict__.keys():
            if time_type in self.times_running:
                logging.warning(f"Overwriting unfinished timing of {time_type}!")
            self.times_running[time_type] = current_time()

    def _archive_timing(self, time_type, duration):
        if time_type not in self.times_archive:
            self.times_archive[time_type] = []
        self.times_archive[time_type].extend([duration])

    def end(self, time_type, gpu=None):
        if time_type in self.times_running:
            duration = current_time() - self.times_running.pop(time_type)
            self._archive_timing(time_type, duration=duration)
            if self.print_all_single_time_stats:
                logging.info(f"{time_type:<10}: took {duration}")
        else:
            logging.warning(f"Attempting to finish timer type {time_type}, which was never started!")

    def complete_all(self):
        pass

    def print_all_time_stats(self):
        for _type, times in self.times_archive.items():
            print(
                f"{_type:<15}; "
                f"min: {min(times):.04e}, "
                f"max: {max(times):.04e}, "
                f"mean: {np.mean(times):.04e}, "
                f"median: {np.median(times):.04e}, "
                f"counts: {len(times):>5}"
            )
            if self.debug:
                for _type, elapsed_time in self.times_running.items():
                    logging.info(f"{_type} still running for {elapsed_time}s.")


def get_max_memory_usage():
    """Max RAM memory used in bytes"""
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1000


class CudaMemoryMonitor:
    def __init__(self) -> None:
        self.init_cuda()

    def init_cuda(self):
        if torch.cuda.is_available():
            torch.cuda.init()
        else:
            logging.info("Cuda initialization failed, as cuda is not available.")

    def reset_cuda_memory_monitoring(self):
        for i_cuda in range(torch.cuda.device_count()):
            torch.cuda.reset_peak_memory_stats(i_cuda)

    def get_cuda_memory_usage(self, log_message, style="verbose"):
        logging.info("CUDA memory logging....\n")
        for i_cuda in range(torch.cuda.device_count()):
            logging.info(f"{log_message}: Cuda device {i_cuda} report:\n")
            if style == "verbose":
                pprint.pprint(torch.cuda.memory_stats(i_cuda))
                logging.info(f"Report done! for Cuda device {i_cuda}\n")
            logging.info(f"Gpu memory currently allocated: {torch.cuda.memory_allocated(i_cuda)/1e9} GB.")
            logging.info(f"Gpu max memory allocated: {torch.cuda.max_memory_allocated(i_cuda)/1e9} GB.")
            logging.info(f"Gpu memory currently reserved: {torch.cuda.memory_reserved(i_cuda)/1e9} GB.")
            logging.info(f"Gpu max memory reserved: {torch.cuda.max_memory_reserved(i_cuda)/1e9} GB.")



def n_model_parameters(model):
    n_trainable = sum(p.numel() for p in model.parameters() if p.requires_grad)
    n_not_trainable = sum(p.numel() for p in model.parameters() if not p.requires_grad)
    n_total = sum(p.numel() for p in model.parameters())
    n_trainable_plus_not_trainable = n_trainable + n_not_trainable
    return {
        "n_trainable": n_trainable,
        "n_not_trainable": n_not_trainable,
        "n_total": n_total,
        "n_trainable_plus_not_trainable": n_trainable_plus_not_trainable,
    }



import ast
import collections
import datetime
import functools
import itertools
import logging
import multiprocessing
import pathlib
import re
import typing as t
import warnings
from collections.abc import Iterable
from functools import wraps
from time import time
from typing import Any
from typing import Optional

import numpy as np
import pandas as pd


def timing(f: t.Callable):
    """
    Wrapper that returns execution time and arguments of function.
    """

    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logging.info(f"func:{f.__name__} args:[{kw}] took: {te-ts} sec")
        return result

    return wrap


def get_processes(processes):
    if processes < 0:
        processes = multiprocessing.cpu_count() + processes + 1
    return processes


def parallelize(
    function: t.Callable,
    args_zipped: t.Iterable,
    processes: int = -1,
    single_arg: bool = False,
    kwargs_as_dict: t.Optional[dict] = None,
):
    """
    parallelize function with args provided in zipped format
    ----------
    function: function to parallelize
    args: args of function in zipped format

    Returns
    -------
    function applied to args
    """
    if processes == 1:
        results = []
        if kwargs_as_dict is None:
            kwargs_as_dict = {}
        for args in args_zipped:
            if isinstance(args, str):
                args = (args,)
            results.append(function(*args, **kwargs_as_dict))
        return results
    processes = get_processes(processes)
    print(f"n processes: {processes}")
    if single_arg and kwargs_as_dict is None:
        with multiprocessing.Pool(processes=processes) as pool:
            results = pool.map(function, args_zipped)
    else:
        if single_arg:
            args_zipped = ((arg,) for arg in args_zipped)
        if kwargs_as_dict is not None:
            kwargs_iter = itertools.repeat(kwargs_as_dict)
        else:
            kwargs_iter = itertools.repeat(dict())
        with multiprocessing.Pool(processes=processes) as pool:
            results = starmap_with_kwargs(pool, function, args_zipped, kwargs_iter)
    return results


def get_random_indices(n_sample: int, size_data: int) -> np.ndarray:
    """Draw `n_sample` indices from range(`size_data`)"""
    return np.random.choice(
        range(size_data),
        n_sample if n_sample < size_data else size_data,
        replace=False,
    )


def starmap_with_kwargs(pool, function: t.Callable, args_iter: t.Iterable, kwargs_iter: t.Iterable):
    """Helper function to parallelize functions with args and kwargs"""
    if kwargs_iter is None:
        args_for_starmap = zip(itertools.repeat(function), args_iter)
    else:
        args_for_starmap = zip(itertools.repeat(function), args_iter, kwargs_iter)
    return pool.starmap(apply_args_and_kwargs, args_for_starmap)


def apply_args_and_kwargs(fn: t.Callable, args, kwargs):
    """Helper function to parallelize functions with args and kwargs"""
    return fn(*args, **kwargs)


def evaluate_string(x: object):
    try:
        return ast.literal_eval(str(x))
    except ValueError:
        return x


def chunks(sequence: t.Sequence, n: int) -> t.Iterator:
    """Yield successive `n`-sized chunks from `sequence`."""
    if n == -1:
        n = len(sequence)
    for i in range(0, len(sequence), n):
        yield sequence[slice(i, i + n)]


def round_to_base(x: float, base: float):
    return base * round(x / base)


def round_offset(values: np.ndarray, decimal: int, offset: float) -> np.ndarray:
    values += offset
    values = np.round(values, decimals=decimal)
    values -= offset
    return values


def round_numpy_time_to_base_minutes(time: np.datetime64, base: int = 5) -> np.datetime64:
    tm = datetime.datetime.utcfromtimestamp(time.tolist() / 1e9)
    tm += datetime.timedelta(minutes=base / 2)
    tm -= datetime.timedelta(minutes=tm.minute % base, seconds=tm.second, microseconds=tm.microsecond)
    return np.datetime64(tm)


def round_time_to_base_minutes(time: np.datetime64, base: int = 5):
    tm = pd.to_datetime(time).round(f"{base}min")
    return np.datetime64(tm)


def vetorize_time_to_base_minutes(times: np.ndarray, base: int = 5):
    function = functools.partial(round_numpy_time_to_base_minutes, base=base)
    rounded = np.array(list(map(function, times)))
    return rounded


def _assert_same_type(variable, type_from_variable):
    """Check if `variable` shares type with `type_from_variable`"""
    if isinstance(variable, collections.abc.Iterable):
        if not isinstance(type_from_variable, collections.abc.Iterable):
            raise TypeError(f"Variable {variable=} is iterable but type is not {type_from_variable=}")
        for v, typ in zip(variable, type_from_variable):
            _assert_same_type(v, typ)
    else:
        if not isinstance(variable, type(type_from_variable)):
            raise TypeError(f"Variable {variable=} different type than expected {type_from_variable=}")


def assert_same_type_as(variable: object, type_from_variable: object, alternative: Optional[object] = None):
    """Check if `variable` is same type as `type_from_variable` unless `variable` is `alternative`"""
    if variable is alternative:
        return
    try:
        _assert_same_type(variable, type_from_variable)
    except TypeError:
        raise TypeError(f"Variable {variable} not of expected type {type_from_variable}")


def all_same_type(variable_list: t.Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def assert_shape(variable, shape: t.Tuple, name: Optional[str] = None, ignore_none: bool = True):
    if ignore_none:
        if variable is None:
            return
    if np.shape(variable) != shape:
        raise TypeError(f"{name+': ' if name is not None else ''}{variable=} doesn't have required {shape=}!")


def flatten_list(lis: list[list]) -> list:
    return [item for sublist in lis for item in sublist]


def str_to_delta_time(string: str) -> t.Tuple[float, str]:
    time, units, _ = re.split("([a-zA-Z]+)$", string)
    return float(time), units


def to_nlength_tuple(x: object, n: int = 2) -> tuple:
    if not (isinstance(x, list) or isinstance(x, tuple) or isinstance(x, np.ndarray)):
        return tuple([x] * n)
    elif isinstance(x, list) or isinstance(x, np.ndarray):
        x = tuple(x)
    if len(x) != n:
        raise ValueError(f"{x} doesn't have expected length {n=}")
    return x


def assert_all_same_type(variable_list: Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def validate_array(array: np.ndarray, type_: str = "float", name: str | None = None):
    """Validate numpy based on `type`, e.g. no nan-values"""
    if type_ == "float":
        if np.sum(np.isnan(array.astype(float))) > 0:
            raise ValueError(f"Found nan-values {f'in {name}' if name is not None else ''}!")
    else:
        raise ValueError(f"{type_=} unknown!")


def override_class_method(class_instance, method_name, target_class):
    class_method = getattr(target_class, method_name)

    def new_method(*args, **kwargs):
        return class_method(class_instance, *args, **kwargs)

    setattr(class_instance, method_name, new_method)


def is_in_list_and_remove(to_check, _list):
    is_in = False
    if to_check in _list:
        is_in = True
    if is_in:
        _list.remove(to_check)
    return is_in


def _import_torch(file):
    try:
        import torch as torch  # noqa: E501
    except ModuleNotFoundError:
        warnings.warn(f"Need to install `torch` to use all functionality in {pathlib.Path(file).parent}.")
    else:
        return torch


def _import_xarray_and_define_xarray_type(file, also_return_dataarray=False):
    try:
        import xarray as xarray

        xarray_dataset_type = xarray.Dataset
        xarray_dataarray_type = xarray.DataArray
    except ImportError as e:
        xarray_dataset_type = Any
        xarray_dataarray_type = Any
        logging.warn(f"xarray couldn't be imported to use all functionality in {pathlib.Path(file).parent}.\n{e}")
        xarray = None
    if also_return_dataarray:
        return xarray, xarray_dataset_type, xarray_dataarray_type
    return xarray, xarray_dataset_type



import logging
import os
import re
import typing as t
from typing import Optional
from typing import Type

import a2.dataset
import a2.utils.constants
import a2.utils.testing
import a2.utils.utils
import numpy as np
import pandas as pd

xarray, xarray_dataset_type, xarray_dataarray_type = a2.utils.utils._import_xarray_and_define_xarray_type(
    __file__, also_return_dataarray=True
)


def _get_dataset_backend():
    backend = os.environ.get("A2_DATASET_BACKEND", "xarray")
    if backend not in a2.utils.constants.TYPE_DATASET_BACKEND:
        raise NotImplementedError(f"{backend=} not recognized!")
    return backend


def _using_xarray():
    return "xarray" == _get_dataset_backend()


def _using_pandas():
    return "pandas" == _get_dataset_backend()


def is_same_type_data_array(ds: xarray_dataset_type, field: t.Hashable, which_type: type = str):
    return all(isinstance(x, which_type) for x in ds[field].values)


def is_nan(
    ds: xarray_dataset_type,
    field: t.Union[str, t.Hashable],
    dims: Optional[t.Tuple] = None,
):
    """
    Test element-wise for nan-values and returns boolean mask as DataArray

    If DataArray consists only of strings, "nan" is assumed to be a nan-value.
    Parameters:
    ----------
    ds: dataset
    field: Key variable to check for nans
    dims: Dimensions of field

    Returns
    -------
    Boolean DataArray
    """
    if dims is None:
        dims = ds[field].coords.dims
    if is_same_type_data_array(ds, field):
        return xarray.DataArray(ds[field].values == "nan", dims=dims)
    else:
        return xarray.DataArray(pd.isna(ds[field].values), dims=dims)


def ds_shape(ds, variable=None):
    if _using_xarray():
        if variable is None:
            variable = "index"
        shape = np.shape(ds[variable])
    elif _using_pandas():
        shape = ds.shape
    return shape


def drop_nan(
    ds,
    columns,
    dims: Optional[t.Tuple] = None,
):
    if _using_xarray():

        def _compute_mask(ds, column, dims):
            if dims is None:
                dims = ds[column].coords.dims
            return xarray.DataArray(~is_nan(ds, column).values, dims=dims)

        mask = _compute_mask(ds, columns.pop(0), dims)
        for c in columns:
            mask = np.logical_and(mask, _compute_mask(ds, c, dims))
        logging.info(f"Dropping {np.sum(~mask.values)}/{len(ds.index)} nan-values.")
        return ds.where(mask, drop=True)
    elif _using_pandas():
        return ds.dropna(subset=columns)


def is_na(
    ds: xarray_dataset_type,
    field: str,
    check: t.Optional[t.List[str]] = None,
    dims: Optional[t.Tuple] = None,
):
    """
    Returns boolean data array that notes if either field in `check` is matched.

    Parameters:
    ----------
    ds: Dataset
    field: Key variable to check for na values
    check: Values to check for that are considered 'na'
    dims: Dimensions of field

    Returns
    -------
    dict of json response
    """
    if dims is None:
        dims = ds[field].coords.dims
    if check is None:
        check = ["nan", ""]
    conditions = []
    if "nan" in check:
        conditions.append(is_nan(ds, field, dims=dims))
        check.remove("nan")
    for c in check:
        conditions.append(xarray.DataArray(ds[field] == c, dims=dims))
    combine = conditions[0]
    for cond in conditions:
        combine = np.logical_or(combine, cond)
    return combine


def print_tweet_sample(
    ds: xarray_dataset_type,
    n_sample: int = 5,
    n: int | None = None,
    fancy: bool = True,
    field_name_tweets: str = "text",
    additional_fields: list | None = [],
):
    """load tweets in filename and print random n_sample or n tweets from
    beginning of file"""
    fields = [field_name_tweets]
    if additional_fields is not None:
        fields += additional_fields
    print_sample(np.array([ds[key].values for key in fields]).T, n_sample=n_sample, n=n, fancy=fancy)


def print_sample(
    data: t.Union[t.Sequence, np.ndarray],
    n_sample: int = 5,
    n: t.Optional[int] = None,
    fancy: bool = True,
):
    """print data sample of sequence"""
    size_data = len(data)
    index_list = get_sample_indices(n, n_sample, size_data)
    for i in index_list:
        if fancy:
            print("------------------------------")
        print(data[i])


def get_sample_indices(n: t.Optional[int], n_sample: int, size_data: int) -> list:
    """Draws sample indices:
    n is None: returns `n_sample` randomly drawn values from size_data
    n < 0: returns range(size_data)
    else: returns range(n)
    """
    if n is None:
        index_list = list(a2.utils.utils.get_random_indices(n_sample=n_sample, size_data=size_data))
    elif n < 0:
        index_list = [i for i in range(size_data)]
    else:
        index_list = [i for i in range(n if n < size_data else size_data)]
    return index_list


def print_tweet_groupby(
    ds: xarray_dataset_type,
    group_by: t.Union[str, xarray_dataarray_type],
    n_sample: int = 5,
    n: Optional[int] = None,
    n_groups: int = 20,
    ds_grouped: Optional[xarray_dataset_type] = None,
    fancy: bool = True,
    fields_to_print: list = ["text"],
) -> xarray_dataset_type:
    """
    Print example tweets created per authors sorted by most active
    authors and source of post.

    Necessary grouping and summing steps are very costly such that the
    resulting dataset of this operation is returned
    by this function and can be provided as an argument `authors` to
    speed up additional prints

    Example:
        group_by_source = a2.dataset.utils_dataset.print_tweet_groupby(
            ds_tweets, "source", n_groups=10
        )
    Parameters:
    ----------
    ds: dataset
    n_sample: size of randomly selected tweets to printed per author
    n: number of tweets printed as ordered in dataset
    n_groups: number of authors shown
    authors: return value of this function, when provided for additional
             printing of same dataset, speeds up process considerably

    Returns
    -------
    dict of json response
    """
    ds_grouped, sort_by = dataset_groupby(ds, group_by, ds_grouped)
    group_by_variable_name = list(ds_grouped.coords.variables.keys())[0]
    for i, (number_tweets_per_group, group_name) in enumerate(
        zip(
            ds_grouped[sort_by].values,
            ds_grouped[group_by_variable_name].values,
        )
    ):
        if i >= n_groups:
            break
        if fancy:
            print("------------------------------")
        print(f"{group_name} --> {number_tweets_per_group}")
        if isinstance(group_by, str):
            mask = ds[group_by].values == group_name
        elif isinstance(group_by, xarray.DataArray):
            mask = group_by.values == group_name
        else:
            raise ValueError(f"wrong type group_by: {group_by}")
        if "source" in ds:
            unique_sources = np.unique(ds["source"].loc[mask].values)
            source_to_print = unique_sources[0] if len(unique_sources) else unique_sources
            print(f"source: {source_to_print}")
        ds_sample = a2.dataset.load_dataset.reset_index_coordinate(ds).sel(index=mask)
        index_list = get_sample_indices(
            n,
            n_sample,
            ds_sample.index.shape[0] if ds_sample.index.shape else 0,
        )
        info_text = info_tweets_to_text(
            a2.dataset.load_dataset.reset_index_coordinate(ds_sample).sel(index=index_list),
            fields=fields_to_print,
            fancy=fancy,
        )
        print(info_text)
    return ds_grouped


def get_keywords_default():
    header = (
        "🏔️ OR 🏔️ OR ☀️ OR ☀️ OR 🌞 OR ⛅ OR ⛈️ OR ⛈️ OR 🌤️ OR 🌤️ OR 🌥️ OR 🌥️ OR"
        " 🌦️ OR 🌦️ OR 🌧️ OR 🌧️ OR 🌨️ OR 🌨️ OR 🌩️ OR 🌩️ OR ☔ OR ⛄"
        " OR blizzard OR cloudburst OR downpour OR drizzle OR flash"
        " flood OR flood OR flood stage OR forecast OR freezing rain"
        " OR hail OR ice storm OR lightning OR precipitation OR rain"
        " OR rain gauge OR rain shadow OR rainbands OR rain shower OR"
        " snow OR snow shower OR snowstorm OR sun OR sunny OR thunder"
        " OR thunderstorm"
    )
    keywords = header.split(" OR ")
    return keywords


def dataset_groupby(
    ds: xarray_dataset_type,
    group_by: t.Union[str, xarray_dataarray_type],
    ds_grouped: t.Union[xarray_dataset_type, None] = None,
) -> t.Tuple[xarray_dataset_type, t.Hashable]:
    if ds_grouped is None:
        ds_grouped_unsorted = ds.groupby(group_by).count()
        sort_by = get_variable_name_first(ds_grouped_unsorted)
        ds_grouped = ds_grouped_unsorted.sortby(sort_by, ascending=False)
    else:
        sort_by = get_variable_name_first(ds_grouped)
    return ds_grouped, sort_by


def filter_tweets(ds: xarray_dataset_type, terms: list[str], min_tweets_per_author: int = 0):
    """
    Filter out tweets if include terms

    Tweets including any term are filtered out if its author used
    the term in more tweets than `min_tweets_per_author`.
    Parameters:
    ----------
    ds: dataset
    terms: List of strings with terms.
    min_tweets_per_author:

    Returns
    -------
    filtered dataset
    """
    tweet_id_name = "tweet_id" if "tweet_id" in ds.variables else "id"
    initial_n_tweets = ds["index"].shape[0]
    if "author_id" not in ds.variables.keys():
        raise ValueError("Missing variable 'author_id' in ds!")
    for te in terms:
        where = ds.where(ds.text.str.contains(f"{te}", flags=re.IGNORECASE), drop=True)
        if not where["index"].shape[0]:
            continue
        counts = where.groupby("author_id").count().sortby(tweet_id_name, ascending=False)
        logging.info("------------------------")
        logging.info(
            f"found {np.sum(counts[tweet_id_name].values)} tweets from"
            f" {counts[tweet_id_name].shape[0]} author_ids's with term {te}"
        )
        for author, n in zip(counts["author_id"].astype(int).values, counts[tweet_id_name].values):
            sample = np.random.choice(
                ds["text"].loc[ds.author_id.values == author].values,
                10 if n > 10 else n,
                replace=False,
            )
            logging.info("------------------------")
            logging.info(f"{author} posted {n} tweets with term: {te}")
            logging.info(f"sample: {sample}")
            if n > min_tweets_per_author:
                ds = ds.where(~(ds.author_id == author), drop=True)
    final_n_tweets = ds["index"].shape[0]
    logging.info(f"removed {initial_n_tweets-final_n_tweets} tweets")
    return ds


def print_variables(ds: xarray_dataset_type):
    for k, v in ds.variables.items():
        print(f"{k} --> {v}")


def info_tweets_to_text(ds: xarray_dataset_type, fields: t.Optional[t.Sequence[str]] = None, fancy=False):
    if fields is None:
        fields = [
            "text",
            "latitude_rounded",
            "longitude_rounded",
            "created_at",
            "tp",
            "raining",
        ]
    unique_fields = np.intersect1d(list(set(fields)), list(set(ds.variables.keys())))
    to_print = ""
    for tweet in zip(*[ds[f].values for f in unique_fields]):
        if fancy:
            to_print += "------------------------------\n"
        for label, value in zip(unique_fields, tweet):
            to_print += f"{label}: {value}\n"
    return to_print


def add_precipitation_memory_efficient(
    ds_tweets,
    ds_weather_filenames,
    key_longitude="longitude_rounded",
    key_latitude="latitude_rounded",
    key_time_tweets="time_half",
    key_time_precipitation="time_half",
    key_precipitation_tweets="tp_new",
    key_precipitation_precipitation="tp_h",
):
    if not isinstance(ds_weather_filenames, list):
        ds_weather_filenames = [ds_weather_filenames]
    for filename in ds_weather_filenames:
        ds_precipitation = a2.dataset.load_dataset.load_tweets_dataset(filename, raw=True)
        ds_tweets = add_precipitation_to_tweets(
            ds_tweets,
            ds_precipitation,
            key_longitude=key_longitude,
            key_latitude=key_latitude,
            key_time_tweets=key_time_tweets,
            key_time_precipitation=key_time_precipitation,
            key_precipitation_tweets=key_precipitation_tweets,
            key_precipitation_precipitation=key_precipitation_precipitation,
        )
    return ds_tweets


def add_precipitation_to_tweets(
    ds_tweets,
    ds_precipitation,
    key_longitude="longitude_rounded",
    key_latitude="latitude_rounded",
    key_time_tweets="time_half",
    key_time_precipitation="time_half",
    key_precipitation_tweets="tp_new",
    key_precipitation_precipitation="tp_h",
):
    """
    adds precipitation columns to twitter dataset

    Parameters:
    ----------
    ds_twit: twitter dataset
    ds_prec: precipitation dataset
    key_longitude: name of longitude field in both datasets
    key_latitude: name of latitude field in both datasets
    key_time_tweets: name of time field in Tweets datasets
    key_time_precipitation: name of time field in precipitation datasets
    key_precipitation_tweets: name of precipitation field in Tweets dataset
    key_precipitation_precipitation: name of precipitation field in
                                     precipitation datasets

    Returns
    -------
    twitter dataset with precipitation column added
    """
    a2.utils.testing.assert_presence_variables(ds_tweets, ["longitude", "latitude", "created_at"])
    a2.utils.testing.assert_presence_variables(
        ds_precipitation,
        ["longitude", "latitude", key_precipitation_precipitation],
    )
    if key_precipitation_tweets not in ds_tweets.variables.keys():
        ds_tweets = a2.dataset.utils_dataset.initialize_variable(ds_tweets, key_precipitation_tweets, dtype=object)
    if "time" not in ds_tweets.variables.keys():
        ds_tweets["time"] = (["index"], ds_tweets.created_at.values)
    ds_tweets = add_field(ds_tweets, key_time_tweets, coordinates=["index"])
    ds_tweets = add_field(ds_tweets, key_longitude, coordinates=["index"])
    ds_tweets = add_field(ds_tweets, key_latitude, coordinates=["index"])

    ds_precipitation = add_field(ds_precipitation, key_time_precipitation)
    ds_precipitation = add_field(ds_precipitation, key_longitude, rename_coordinate="longitude")
    ds_precipitation = add_field(ds_precipitation, key_latitude, rename_coordinate="latitude")
    mask = (ds_tweets[key_time_tweets] >= ds_precipitation[key_time_tweets].values.min()) & (
        ds_tweets[key_time_tweets] <= ds_precipitation[key_time_tweets].values.max()
    )
    time = ds_tweets[key_time_tweets].loc[mask]
    longitude = ds_tweets[key_longitude].loc[mask]
    latitude = ds_tweets[key_latitude].loc[mask]
    ds_tweets[key_precipitation_tweets].loc[mask] = ds_precipitation.sel(
        {
            key_longitude: longitude,
            key_latitude: latitude,
            key_time_precipitation: time,
        }
    )[key_precipitation_precipitation].values
    number_nan_values = np.sum(a2.dataset.utils_dataset.is_nan(ds_tweets, key_precipitation_tweets))
    logging.info(f"found {number_nan_values} nan-values in precipitation dataset")
    return ds_tweets


def add_field(
    ds: xarray_dataset_type,
    variable: t.Hashable,
    coordinates: Optional[list] = None,
    overwrite: bool = False,
    rename_coordinate: Optional[str] = None,
):
    if coordinates is None:
        coordinates = list(ds.coords)
    if variable in ds and not overwrite:
        return ds
    if rename_coordinate is not None and not isinstance(rename_coordinate, str):
        raise Exception(f"rename_coordinate: {rename_coordinate} should be field name of " "dataset or None!")
    if variable == "longitude_rounded":
        values = np.round(ds["longitude"].astype(float).values, decimals=1)
    elif variable == "latitude_rounded":
        values = np.round(ds["latitude"].astype(float).values, decimals=1)
    elif variable == "time_half":
        values = pd.to_datetime(ds.time.values + pd.Timedelta("30min")).round("1h") - pd.Timedelta("30min")
    elif variable == "time_h":
        values = pd.to_datetime(ds.time).round("1h").values
    else:
        raise ValueError(f"Variable {variable} not implemented!")

    if rename_coordinate is not None:
        ds = ds.rename({rename_coordinate: variable})
        ds[rename_coordinate] = (
            variable,
            ds[variable].values,
        )
        ds = add_coordinates(ds, variable, values)
    else:
        ds[variable] = (coordinates, values)

    return ds


def add_coordinates(ds: xarray_dataset_type, key: t.Hashable, values: np.ndarray) -> xarray_dataset_type:
    ds[key] = values
    return ds


def add_variable(
    ds: xarray_dataset_type,
    key: t.Hashable,
    values: np.ndarray,
    coordinate: list | None = None,
    type_values: Type | None = None,
) -> xarray_dataset_type:
    if type_values is None:
        type_values = values.dtype
    if _using_xarray():
        if coordinate is None:
            coordinate = ["index"]
        ds[key] = (coordinate, np.array(values, dtype=type_values))
    elif _using_pandas():
        ds[key] = np.array(values, dtype=type_values)
    return ds


def select_rows_by_index(
    ds: xarray_dataset_type,
    indices,
):
    if _using_xarray():
        ds = ds.sel(index=indices)
    elif _using_pandas():
        ds = ds.iloc[indices]
    return ds


def drop_rows(
    ds: xarray_dataset_type,
    condition: np.ndarray,
) -> xarray_dataset_type:
    if _using_xarray():
        ds = ds.where(condition, drop=True)
    elif _using_pandas():
        ds = ds.loc[condition]
    return ds


def initialize_variable(
    ds: xarray_dataset_type,
    field: str,
    coordinates=["index"],
    values=np.nan,
    dtype: type = float,
) -> xarray_dataset_type:
    variable_name_first = get_variable_name_first(ds)
    ds[field] = (
        coordinates,
        np.full(ds[variable_name_first].values.shape[0], values, dtype=dtype),
    )
    return ds


def get_variable_name_first(ds: xarray_dataset_type) -> t.Hashable:
    if _using_xarray():
        first_variable = list(ds.variables.keys())[0]
    elif _using_pandas():
        first_variable = ds.columns[0]
    return first_variable


def array_elements_to_str(array: np.ndarray) -> np.ndarray:
    return np.array(str(x) for x in array)


def assert_keys_in_dataset(ds, keys):
    if _using_xarray():
        variables = ds.variables.keys()
    elif _using_pandas():
        variables = ds.columns
    not_found = [k for k in keys if k not in variables]
    if not not_found:
        return True
    else:
        raise ValueError(f"Couldn't find keys {not_found}!")


def divide_ds_by_unique_values(ds, key_divide_by, keys_values):
    ds = ds.sortby(key_divide_by)
    ds = a2.dataset.load_dataset.reset_index_coordinate(ds)
    _, indices, counts = np.unique(ds[key_divide_by].values.astype(int), return_index=True, return_counts=True)
    index_list = [[i for _ in range(c)] for i, c in zip(indices, counts)]
    divided_values = []
    for k in keys_values:
        divided_values.append(
            [
                [ds[k].values[i + index_inner] for index_inner, i in enumerate(l)]
                for index_outer, l in enumerate(index_list)
            ]
        )
    return ds, divided_values


def construct_dataset(ds, data_vars, time=None):
    coords = ds.coords
    if time is not None:
        coords["time"] = time
    return xarray_dataset_type(data_vars=data_vars, coords=coords)


def merge_datasets_along_index(ds_top: xarray_dataset_type, ds_bottom: xarray_dataset_type) -> xarray_dataset_type:
    ds_bottom_reindexed = ds_bottom.copy()
    ds_top = ds_top.copy()

    ds_top = a2.dataset.load_dataset.reset_index_coordinate(ds_top)
    ds_bottom_reindexed = a2.dataset.load_dataset.reset_index_coordinate(ds_bottom_reindexed)

    start_index = ds_top.index.shape[0]
    ds_bottom_reindexed["index"] = range(start_index, start_index + ds_bottom_reindexed.index.shape[0])
    return xarray.merge([ds_top, ds_bottom_reindexed])
